@@ -13,7 +13,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 #include <stdlib.h>
 #include <string.h>
@@ -42,10 +42,11 @@ PaintCore  non_gui_paint_core;
 static MaskBuf * paint_core_subsample_mask  (MaskBuf *, double, double);
 static MaskBuf * paint_core_solidify_mask   (MaskBuf *);
 static MaskBuf * paint_core_get_brush_mask  (PaintCore *, int);
-static void      paint_core_paste           (PaintCore *, MaskBuf *, int, int, int, int, int);
+static void      paint_core_paste           (PaintCore *, MaskBuf *, GimpDrawable *, int, int, int, int);
+static void      paint_core_replace         (PaintCore *, MaskBuf *, GimpDrawable *, int, int, int);
 static void      paint_to_canvas_tiles      (PaintCore *, MaskBuf *, int);
 static void      paint_to_canvas_buf        (PaintCore *, MaskBuf *, int);
-static void      set_undo_tiles             (int, int, int, int, int);
+static void      set_undo_tiles             (GimpDrawable *, int, int, int, int);
 static void      set_canvas_tiles           (int, int, int, int);
 
 
@@ -67,10 +68,7 @@ static TempBuf *  canvas_buf = NULL;
 
 /*  brush buffers  */
 static MaskBuf *  solid_brush;
-static MaskBuf *  kernel_brushes[4][4] = { {NULL, NULL, NULL, NULL},
-					   {NULL, NULL, NULL, NULL},
-					   {NULL, NULL, NULL, NULL},
-					   {NULL, NULL, NULL, NULL}};
+static MaskBuf *  kernel_brushes[5][5];
 
 
 /*  paint buffers utility functions  */
@@ -84,32 +82,42 @@ static void        free_paint_buffers     (void);
 #define KERNEL_HEIGHT  3
 
 /*  Brush pixel subsampling kernels  */
-static int subsample[4][4][9] =
-{
-  {
-    { 16, 48, 0, 48, 144, 0, 0, 0, 0 },
-    { 0, 64, 0, 0, 192, 0, 0, 0, 0 },
-    { 0, 48, 16, 0, 144, 48, 0, 0, 0 },
-    { 0, 32, 32, 0, 96, 96, 0, 0, 0 },
-  },
-  {
-    { 0, 0, 0, 64, 192, 0, 0, 0, 0 },
-    { 0, 0, 0, 0, 256, 0, 0, 0, 0 },
-    { 0, 0, 0, 0, 192, 64, 0, 0, 0 },
-    { 0, 0, 0, 0, 128, 128, 0, 0, 0 },
-  },
-  {
-    { 0, 0, 0, 48, 144, 0, 16, 48, 0 },
-    { 0, 0, 0, 0, 192, 0, 0, 64, 0 },
-    { 0, 0, 0, 0, 144, 48, 0, 48, 16 },
-    { 0, 0, 0, 0, 96, 96, 0, 32, 32 },
-  },
-  {
-    { 0, 0, 0, 32, 96, 0, 32, 96, 0 },
-    { 0, 0, 0, 0, 128, 0, 0, 128, 0 },
-    { 0, 0, 0, 0, 96, 32, 0, 96, 32 },
-    { 0, 0, 0, 0, 64, 64, 0, 64, 64 },
-  },
+static int subsample[5][5][9] = {
+	{
+		{ 64, 64, 0, 64, 64, 0, 0, 0, 0, },
+		{ 32, 96, 0, 32, 96, 0, 0, 0, 0, },
+		{ 0, 128, 0, 0, 128, 0, 0, 0, 0, },
+		{ 0, 96, 32, 0, 96, 32, 0, 0, 0, },
+		{ 0, 64, 64, 0, 64, 64, 0, 0, 0, },
+	},
+	{
+		{ 32, 32, 0, 96, 96, 0, 0, 0, 0, },
+		{ 16, 48, 0, 48, 144, 0, 0, 0, 0, },
+		{ 0, 64, 0, 0, 192, 0, 0, 0, 0, },
+		{ 0, 48, 16, 0, 144, 48, 0, 0, 0, },
+		{ 0, 32, 32, 0, 96, 96, 0, 0, 0, },
+	},
+	{
+		{ 0, 0, 0, 128, 128, 0, 0, 0, 0, },
+		{ 0, 0, 0, 64, 192, 0, 0, 0, 0, },
+		{ 0, 0, 0, 0, 256, 0, 0, 0, 0, },
+		{ 0, 0, 0, 0, 192, 64, 0, 0, 0, },
+		{ 0, 0, 0, 0, 128, 128, 0, 0, 0, },
+	},
+	{
+		{ 0, 0, 0, 96, 96, 0, 32, 32, 0, },
+		{ 0, 0, 0, 48, 144, 0, 16, 48, 0, },
+		{ 0, 0, 0, 0, 192, 0, 0, 64, 0, },
+		{ 0, 0, 0, 0, 144, 48, 0, 48, 16, },
+		{ 0, 0, 0, 0, 96, 96, 0, 32, 32, },
+	},
+	{
+		{ 0, 0, 0, 64, 64, 0, 64, 64, 0, },
+		{ 0, 0, 0, 32, 96, 0, 32, 96, 0, },
+		{ 0, 0, 0, 0, 128, 0, 0, 128, 0, },
+		{ 0, 0, 0, 0, 96, 32, 0, 96, 32, },
+		{ 0, 0, 0, 0, 64, 64, 0, 64, 64, },
+	},
 };
 
 void
@@ -120,17 +128,17 @@ paint_core_button_press (tool, bevent, gdisp_ptr)
 {
   PaintCore * paint_core;
   GDisplay * gdisp;
-  int drawable_id;
   int draw_line = 0;
   double x, y;
+  GimpDrawable *drawable;
 
   gdisp = (GDisplay *) gdisp_ptr;
   paint_core = (PaintCore *) tool->private;
 
   gdisplay_untransform_coords_f (gdisp, (double) bevent->x, (double) bevent->y, &x, &y, TRUE);
-  drawable_id = gimage_active_drawable (gdisp->gimage);
+  drawable = gimage_active_drawable (gdisp->gimage);
 
-  if (! paint_core_init (paint_core, drawable_id, x, y))
+  if (! paint_core_init (paint_core, drawable, x, y))
     return;
 
   paint_core->state = bevent->state;
@@ -160,22 +168,28 @@ paint_core_button_press (tool, bevent, gdisp_ptr)
   /*  pause the current selection and grab the pointer  */
   gdisplays_selection_visibility (gdisp->gimage->ID, SelectionPause);
 
-  gdk_pointer_grab (gdisp->canvas->window, FALSE,
-		    GDK_POINTER_MOTION_HINT_MASK | GDK_BUTTON1_MOTION_MASK | GDK_BUTTON_RELEASE_MASK,
-		    NULL, NULL, bevent->time);
-
+  /* add motion memory if you press mod1 first */
+  if (bevent->state & GDK_MOD1_MASK)
+    gdk_pointer_grab (gdisp->canvas->window, FALSE,
+		      GDK_BUTTON1_MOTION_MASK | GDK_BUTTON_RELEASE_MASK,
+		      NULL, NULL, bevent->time);
+  else
+    gdk_pointer_grab (gdisp->canvas->window, FALSE,
+		      GDK_POINTER_MOTION_HINT_MASK | GDK_BUTTON1_MOTION_MASK | GDK_BUTTON_RELEASE_MASK,
+		      NULL, NULL, bevent->time);
+  
   /*  Let the specific painting function initialize itself  */
-  (* paint_core->paint_func) (paint_core, drawable_id, INIT_PAINT);
+  (* paint_core->paint_func) (paint_core, drawable, INIT_PAINT);
 
   /*  Paint to the image  */
   if (draw_line)
     {
-      paint_core_interpolate (paint_core, drawable_id);
+      paint_core_interpolate (paint_core, drawable);
       paint_core->lastx = paint_core->curx;
       paint_core->lasty = paint_core->cury;
     }
   else
-    (* paint_core->paint_func) (paint_core, drawable_id, MOTION_PAINT);
+    (* paint_core->paint_func) (paint_core, drawable, MOTION_PAINT);
 
   gdisplay_flush (gdisp);
 }
@@ -248,10 +262,13 @@ paint_core_cursor_update (tool, mevent, gdisp_ptr)
   gdisp = (GDisplay *) gdisp_ptr;
 
   gdisplay_untransform_coords (gdisp, mevent->x, mevent->y, &x, &y, FALSE, FALSE);
-  if ((layer = gimage_get_active_layer (gdisp->gimage)))
-    if (x >= layer->offset_x && y >= layer->offset_y &&
-	x < (layer->offset_x + layer->width) &&
-	y < (layer->offset_y + layer->height))
+  if ((layer = gimage_get_active_layer (gdisp->gimage))) 
+    {
+      int off_x, off_y;
+      drawable_offsets (GIMP_DRAWABLE(layer), &off_x, &off_y);
+    if (x >= off_x && y >= off_y &&
+	x < (off_x + drawable_width (GIMP_DRAWABLE(layer))) &&
+	y < (off_y + drawable_height (GIMP_DRAWABLE(layer))))
       {
 	/*  One more test--is there a selected region?
 	 *  if so, is cursor inside?
@@ -261,6 +278,7 @@ paint_core_cursor_update (tool, mevent, gdisp_ptr)
 	else if (gimage_mask_value (gdisp->gimage, x, y))
 	  ctype = GDK_PENCIL;
       }
+    }
   gdisplay_install_tool_cursor (gdisp, ctype);
 }
 
@@ -272,11 +290,11 @@ paint_core_control (tool, action, gdisp_ptr)
 {
   PaintCore * paint_core;
   GDisplay *gdisp;
-  int drawable_id;
+  GimpDrawable *drawable;
 
   gdisp = (GDisplay *) gdisp_ptr;
   paint_core = (PaintCore *) tool->private;
-  drawable_id = gimage_active_drawable (gdisp->gimage);
+  drawable = gimage_active_drawable (gdisp->gimage);
 
   switch (action)
     {
@@ -287,7 +305,7 @@ paint_core_control (tool, action, gdisp_ptr)
       draw_core_resume (paint_core->core, tool);
       break;
     case HALT :
-      (* paint_core->paint_func) (paint_core, drawable_id, FINISH_PAINT);
+      (* paint_core->paint_func) (paint_core, drawable, FINISH_PAINT);
       paint_core_cleanup ();
       break;
     }
@@ -318,6 +336,7 @@ paint_core_new (type)
   tool->auto_snap_to = TRUE;
   tool->gdisp_ptr = NULL;
   tool->private = (void *) private;
+  tool->preserve = TRUE;
 
   tool->button_press_func = paint_core_button_press;
   tool->button_release_func = paint_core_button_release;
@@ -353,9 +372,9 @@ paint_core_free (tool)
 }
 
 int
-paint_core_init (paint_core, drawable_id, x, y)
+paint_core_init (paint_core, drawable, x, y)
      PaintCore *paint_core;
-     int drawable_id;
+     GimpDrawable *drawable;
      double x, y;
 {
   GBrushP brush;
@@ -366,7 +385,7 @@ paint_core_init (paint_core, drawable_id, x, y)
   /*  Each buffer is the same size as the maximum bounds of the active brush... */
   if (!(brush = get_active_brush ()))
     {
-      warning ("No brushes available for use with this tool.");
+      g_message ("No brushes available for use with this tool.");
       return FALSE;
     }
 
@@ -384,13 +403,13 @@ paint_core_init (paint_core, drawable_id, x, y)
     tile_manager_destroy (canvas_tiles);
 
   /*  Allocate the undo structure  */
-  undo_tiles = tile_manager_new (drawable_width (drawable_id),
-				 drawable_height (drawable_id),
-				 drawable_bytes (drawable_id));
+  undo_tiles = tile_manager_new (drawable_width (drawable),
+				 drawable_height (drawable),
+				 drawable_bytes (drawable));
 
   /*  Allocate the canvas blocks structure  */
-  canvas_tiles = tile_manager_new (drawable_width (drawable_id),
-				   drawable_height (drawable_id), 1);
+  canvas_tiles = tile_manager_new (drawable_width (drawable),
+				   drawable_height (drawable), 1);
 
   /*  Get the initial undo extents  */
   paint_core->x1 = paint_core->x2 = paint_core->curx;
@@ -401,9 +420,9 @@ paint_core_init (paint_core, drawable_id, x, y)
 }
 
 void
-paint_core_interpolate (paint_core, drawable_id)
+paint_core_interpolate (paint_core, drawable)
      PaintCore *paint_core;
-     int drawable_id;
+     GimpDrawable *drawable;
 {
   int n;
   double dx, dy;
@@ -435,7 +454,7 @@ paint_core_interpolate (paint_core, drawable_id)
 
 	  paint_core->curx = paint_core->lastx + dx * t;
 	  paint_core->cury = paint_core->lasty + dy * t;
-	  (* paint_core->paint_func) (paint_core, drawable_id, MOTION_PAINT);
+	  (* paint_core->paint_func) (paint_core, drawable, MOTION_PAINT);
 	}
     }
 
@@ -445,15 +464,15 @@ paint_core_interpolate (paint_core, drawable_id)
 }
 
 void
-paint_core_finish (paint_core, drawable_id, tool_id)
+paint_core_finish (paint_core, drawable, tool_id)
      PaintCore *paint_core;
-     int drawable_id;
+     GimpDrawable *drawable;
      int tool_id;
 {
   GImage *gimage;
   PaintUndo *pu;
 
-  if (! (gimage = drawable_gimage (drawable_id)))
+  if (! (gimage = drawable_gimage (drawable)))
     return;
 
   /*  Determine if any part of the image has been altered--
@@ -473,7 +492,7 @@ paint_core_finish (paint_core, drawable_id, tool_id)
   undo_push_paint (gimage, pu);
 
   /*  push an undo  */
-  drawable_apply_image (drawable_id, paint_core->x1, paint_core->y1,
+  drawable_apply_image (drawable, paint_core->x1, paint_core->y1,
 			paint_core->x2, paint_core->y2, undo_tiles, TRUE);
   undo_tiles = NULL;
 
@@ -483,7 +502,7 @@ paint_core_finish (paint_core, drawable_id, tool_id)
   /*  invalidate the drawable--have to do it here, because
    *  it is not done during the actual painting.
    */
-  drawable_invalidate_preview (drawable_id);
+  drawable_invalidate_preview (drawable);
 }
 
 void
@@ -514,27 +533,27 @@ paint_core_cleanup ()
 /************************/
 
 TempBuf *
-paint_core_get_paint_area (paint_core, drawable_id)
+paint_core_get_paint_area (paint_core, drawable)
      PaintCore *paint_core;
-     int drawable_id;
+     GimpDrawable *drawable;
 {
   int x, y;
   int x1, y1, x2, y2;
   int bytes;
 
-  bytes = drawable_has_alpha (drawable_id) ?
-    drawable_bytes (drawable_id) : drawable_bytes (drawable_id) + 1;
+  bytes = drawable_has_alpha (drawable) ?
+    drawable_bytes (drawable) : drawable_bytes (drawable) + 1;
 
   /*  adjust the x and y coordinates to the upper left corner of the brush  */
   x = (int) paint_core->curx - (paint_core->brush_mask->width >> 1);
   y = (int) paint_core->cury - (paint_core->brush_mask->height >> 1);
 
-  x1 = BOUNDS (x - 1, 0, drawable_width (drawable_id));
-  y1 = BOUNDS (y - 1, 0, drawable_height (drawable_id));
+  x1 = BOUNDS (x - 1, 0, drawable_width (drawable));
+  y1 = BOUNDS (y - 1, 0, drawable_height (drawable));
   x2 = BOUNDS (x + paint_core->brush_mask->width + 1,
-	       0, drawable_width (drawable_id));
+	       0, drawable_width (drawable));
   y2 = BOUNDS (y + paint_core->brush_mask->height + 1,
-	       0, drawable_height (drawable_id));
+	       0, drawable_height (drawable));
 
   /*  configure the canvas buffer  */
   if ((x2 - x1) && (y2 - y1))
@@ -547,9 +566,9 @@ paint_core_get_paint_area (paint_core, drawable_id)
 }
 
 TempBuf *
-paint_core_get_orig_image (paint_core, drawable_id, x1, y1, x2, y2)
+paint_core_get_orig_image (paint_core, drawable, x1, y1, x2, y2)
      PaintCore *paint_core;
-     int drawable_id;
+     GimpDrawable *drawable;
      int x1, y1;
      int x2, y2;
 {
@@ -561,15 +580,15 @@ paint_core_get_orig_image (paint_core, drawable_id, x1, y1, x2, y2)
   unsigned char * s, * d;
   void * pr;
 
-  orig_buf = temp_buf_resize (orig_buf, drawable_bytes (drawable_id),
+  orig_buf = temp_buf_resize (orig_buf, drawable_bytes (drawable),
 			      x1, y1, (x2 - x1), (y2 - y1));
-  x1 = BOUNDS (x1, 0, drawable_width (drawable_id));
-  y1 = BOUNDS (y1, 0, drawable_height (drawable_id));
-  x2 = BOUNDS (x2, 0, drawable_width (drawable_id));
-  y2 = BOUNDS (y2, 0, drawable_height (drawable_id));
+  x1 = BOUNDS (x1, 0, drawable_width (drawable));
+  y1 = BOUNDS (y1, 0, drawable_height (drawable));
+  x2 = BOUNDS (x2, 0, drawable_width (drawable));
+  y2 = BOUNDS (y2, 0, drawable_height (drawable));
 
   /*  configure the pixel regions  */
-  pixel_region_init (&srcPR, drawable_data (drawable_id), x1, y1, (x2 - x1), (y2 - y1), FALSE);
+  pixel_region_init (&srcPR, drawable_data (drawable), x1, y1, (x2 - x1), (y2 - y1), FALSE);
   destPR.bytes = orig_buf->bytes;
   destPR.x = 0; destPR.y = 0;
   destPR.w = (x2 - x1); destPR.h = (y2 - y1);
@@ -612,10 +631,10 @@ paint_core_get_orig_image (paint_core, drawable_id, x1, y1, x2, y2)
 }
 
 void
-paint_core_paste_canvas (paint_core, drawable_id, brush_opacity, image_opacity, paint_mode,
+paint_core_paste_canvas (paint_core, drawable, brush_opacity, image_opacity, paint_mode,
 			 brush_hardness, mode)
      PaintCore *paint_core;
-     int drawable_id;
+     GimpDrawable *drawable;
      int brush_opacity;
      int image_opacity;
      int paint_mode;
@@ -628,10 +647,33 @@ paint_core_paste_canvas (paint_core, drawable_id, brush_opacity, image_opacity,
   brush_mask = paint_core_get_brush_mask (paint_core, brush_hardness);
 
   /*  paste the canvas buf  */
-  paint_core_paste (paint_core, brush_mask, drawable_id,
+  paint_core_paste (paint_core, brush_mask, drawable,
 		    brush_opacity, image_opacity, paint_mode, mode);
 }
 
+/* Similar to paint_core_paste_canvas, but replaces the alpha channel
+   rather than using it to composite (i.e. transparent over opaque
+   becomes transparent rather than opauqe. */
+void
+paint_core_replace_canvas (paint_core, drawable, brush_opacity, image_opacity,
+			   brush_hardness, mode)
+     PaintCore *paint_core;
+     GimpDrawable *drawable;
+     int brush_opacity;
+     int image_opacity;
+     int brush_hardness;
+     int mode;
+{
+  MaskBuf *brush_mask;
+
+  /*  get the brush mask  */
+  brush_mask = paint_core_get_brush_mask (paint_core, brush_hardness);
+
+  /*  paste the canvas buf  */
+  paint_core_replace (paint_core, brush_mask, drawable,
+		      brush_opacity, image_opacity, mode);
+}
+
 /************************************************************
  *             LOCAL FUNCTION DEFINITIONS                   *
  ************************************************************/
@@ -653,22 +695,20 @@ paint_core_subsample_mask (mask, x, y)
   int r, s;
 
   x += (x < 0) ? mask->width : 0;
-  left = x - ((int) x);
+  left = x - floor(x) + 0.125;
   index1 = (int) (left * 4);
-  index1 = (index1 < 0) ? 0 : index1;
 
   y += (y < 0) ? mask->height : 0;
-  left = y - ((int) y);
+  left = y - floor(y) + 0.125;
   index2 = (int) (left * 4);
-  index2 = (index2 < 0) ? 0 : index2;
 
   kernel = subsample[index2][index1];
 
   if ((mask == last_brush) && kernel_brushes[index2][index1])
     return kernel_brushes[index2][index1];
   else if (mask != last_brush)
-    for (i = 0; i < 4; i++)
-      for (j = 0; j < 4; j++)
+    for (i = 0; i < 5; i++)
+      for (j = 0; j < 5; j++)
 	{
 	  if (kernel_brushes[i][j])
 	    mask_buf_free (kernel_brushes[i][j]);
@@ -727,7 +767,7 @@ paint_core_solidify_mask (brush_mask)
       data++;
       for (j = 0; j < brush_mask->width; j++)
 	{
-	  *data++ = (*src++) ? OPAQUE : TRANSPARENT;
+	  *data++ = (*src++) ? OPAQUE_OPACITY : TRANSPARENT_OPACITY;
 	}
       data++;
     }
@@ -750,16 +790,19 @@ paint_core_get_brush_mask (paint_core, brush_hardness)
     case HARD:
       bm = paint_core_solidify_mask (paint_core->brush_mask);
       break;
+    default:
+      bm = NULL;
+      break;
     }
 
   return bm;
 }
 
 static void
-paint_core_paste (paint_core, brush_mask, drawable_id, brush_opacity, image_opacity, paint_mode, mode)
+paint_core_paste (paint_core, brush_mask, drawable, brush_opacity, image_opacity, paint_mode, mode)
      PaintCore *paint_core;
      MaskBuf *brush_mask;
-     int drawable_id;
+     GimpDrawable *drawable;
      int brush_opacity;
      int image_opacity;
      int paint_mode;
@@ -770,11 +813,11 @@ paint_core_paste (paint_core, brush_mask, drawable_id, brush_opacity, image_opac
   TileManager *alt = NULL;
   int offx, offy;
 
-  if (! (gimage = drawable_gimage (drawable_id)))
+  if (! (gimage = drawable_gimage (drawable)))
     return;
 
   /*  set undo blocks  */
-  set_undo_tiles (drawable_id,
+  set_undo_tiles (drawable,
 		  canvas_buf->x, canvas_buf->y,
 		  canvas_buf->width, canvas_buf->height);
 
@@ -805,7 +848,7 @@ paint_core_paste (paint_core, brush_mask, drawable_id, brush_opacity, image_opac
   srcPR.data = temp_buf_data (canvas_buf);
 
   /*  apply the paint area to the gimage  */
-  gimage_apply_image (gimage, drawable_id, &srcPR,
+  gimage_apply_image (gimage, drawable, &srcPR,
 		      FALSE, image_opacity, paint_mode,
 		      alt,  /*  specify an alternative src1  */
 		      canvas_buf->x, canvas_buf->y);
@@ -820,7 +863,86 @@ paint_core_paste (paint_core, brush_mask, drawable_id, brush_opacity, image_opac
    *  instead of drawable_update because we don't want the drawable
    *  preview to be constantly invalidated
    */
-  drawable_offsets (drawable_id, &offx, &offy);
+  drawable_offsets (drawable, &offx, &offy);
+  gdisplays_update_area (gimage->ID, canvas_buf->x + offx, canvas_buf->y + offy,
+			 canvas_buf->width, canvas_buf->height);
+}
+
+/* This works similarly to paint_core_paste. However, instead of combining
+   the canvas to the paint core drawable using one of the combination
+   modes, it uses a "replace" mode (i.e. transparent pixels in the 
+   canvas erase the paint core drawable).
+
+   When not drawing on alpha-enabled images, it just paints using NORMAL
+   mode.
+*/
+static void
+paint_core_replace (paint_core, brush_mask, drawable, brush_opacity, image_opacity, mode)
+     PaintCore *paint_core;
+     MaskBuf *brush_mask;
+     GimpDrawable *drawable;
+     int brush_opacity;
+     int image_opacity;
+     int mode;
+{
+  GImage *gimage;
+  PixelRegion srcPR, maskPR;
+  int offx, offy;
+
+  if (!drawable_has_alpha (drawable))
+    {
+      paint_core_paste (paint_core, brush_mask, drawable,
+			brush_opacity, image_opacity, NORMAL_MODE,
+			mode);
+      return;
+    }
+
+  if (mode != INCREMENTAL)
+    {
+      g_message ("paint_core_replace only works in INCREMENTAL mode");
+      return;
+    }
+
+  if (! (gimage = drawable_gimage (drawable)))
+    return;
+
+  /*  set undo blocks  */
+  set_undo_tiles (drawable,
+		  canvas_buf->x, canvas_buf->y,
+		  canvas_buf->width, canvas_buf->height);
+
+  maskPR.bytes = 1;
+  maskPR.x = 0; maskPR.y = 0;
+  maskPR.w = canvas_buf->width;
+  maskPR.h = canvas_buf->height;
+  maskPR.rowstride = maskPR.bytes * brush_mask->width;
+  maskPR.data = mask_buf_data (brush_mask);
+  
+  /*  intialize canvas buf source pixel regions  */
+  srcPR.bytes = canvas_buf->bytes;
+  srcPR.x = 0; srcPR.y = 0;
+  srcPR.w = canvas_buf->width;
+  srcPR.h = canvas_buf->height;
+  srcPR.rowstride = canvas_buf->width * canvas_buf->bytes;
+  srcPR.data = temp_buf_data (canvas_buf);
+
+  /*  apply the paint area to the gimage  */
+  gimage_replace_image (gimage, drawable, &srcPR,
+			FALSE, image_opacity,
+			&maskPR,
+			canvas_buf->x, canvas_buf->y);
+
+  /*  Update the undo extents  */
+  paint_core->x1 = MINIMUM (paint_core->x1, canvas_buf->x);
+  paint_core->y1 = MINIMUM (paint_core->y1, canvas_buf->y);
+  paint_core->x2 = MAXIMUM (paint_core->x2, (canvas_buf->x + canvas_buf->width));
+  paint_core->y2 = MAXIMUM (paint_core->y2, (canvas_buf->y + canvas_buf->height));
+
+  /*  Update the gimage--it is important to call gdisplays_update_area
+   *  instead of drawable_update because we don't want the drawable
+   *  preview to be constantly invalidated
+   */
+  drawable_offsets (drawable, &offx, &offy);
   gdisplays_update_area (gimage->ID, canvas_buf->x + offx, canvas_buf->y + offy,
 			 canvas_buf->width, canvas_buf->height);
 }
@@ -868,7 +990,7 @@ paint_to_canvas_tiles (paint_core, brush_mask, brush_opacity)
 		     canvas_buf->width, canvas_buf->height, FALSE);
 
   /*  apply the canvas tiles to the canvas buf  */
-  apply_mask_to_region (&srcPR, &maskPR, OPAQUE);
+  apply_mask_to_region (&srcPR, &maskPR, OPAQUE_OPACITY);
 }
 
 static void
@@ -878,6 +1000,14 @@ paint_to_canvas_buf (paint_core, brush_mask, brush_opacity)
      int brush_opacity;
 {
   PixelRegion srcPR, maskPR;
+  int x, y;
+  int xoff, yoff;
+
+  x = (int) paint_core->curx - (brush_mask->width >> 1);
+  y = (int) paint_core->cury - (brush_mask->height >> 1);
+  xoff = (x < 0) ? -x : 0;
+  yoff = (y < 0) ? -y : 0;
+
 
   /*  combine the canvas buf and the brush mask to the canvas buf  */
   srcPR.bytes = canvas_buf->bytes;
@@ -892,15 +1022,15 @@ paint_to_canvas_buf (paint_core, brush_mask, brush_opacity)
   maskPR.w = srcPR.w;
   maskPR.h = srcPR.h;
   maskPR.rowstride = maskPR.bytes * brush_mask->width;
-  maskPR.data = mask_buf_data (brush_mask);
+  maskPR.data = mask_buf_data (brush_mask) + yoff * maskPR.rowstride + xoff * maskPR.bytes;
 
   /*  apply the mask  */
   apply_mask_to_region (&srcPR, &maskPR, brush_opacity);
 }
 
 static void
-set_undo_tiles (drawable_id, x, y, w, h)
-     int drawable_id;
+set_undo_tiles (drawable, x, y, w, h)
+     GimpDrawable *drawable;
      int x, y;
      int w, h;
 {
@@ -915,7 +1045,7 @@ set_undo_tiles (drawable_id, x, y, w, h)
 	  dest_tile = tile_manager_get_tile (undo_tiles, j, i, 0);
 	  if (dest_tile->valid == FALSE)
 	    {
-	      src_tile = tile_manager_get_tile (drawable_data (drawable_id), j, i, 0);
+	      src_tile = tile_manager_get_tile (drawable_data (drawable), j, i, 0);
 	      tile_ref (src_tile);
 	      tile_ref (dest_tile);
 	      memcpy (dest_tile->data, src_tile->data,
